#ifdef TESTS_BUILD

#include <utility>
#include "../../../extern/doctest.hpp"
#include "../../../src/Battling/ScriptHandling/ScriptAggregator.hpp"

using namespace CreatureLib;
using namespace CreatureLib::Battling;

class TestScript : public Script {
private:
    ArbUt::StringView _name;

public:
    explicit TestScript(const std::string& name) : _name(name.c_str(), name.length()){};
    const ArbUt::StringView& GetName() const noexcept override { return _name; }

    void TestMethod(int& runCount) { runCount++; }
};

TEST_CASE("Script Aggregator properly iterates containing script.") {
    auto script = std::make_unique<TestScript>("test");
    auto ran = 0;
    auto vec =
        ArbUt::List<ScriptWrapper>{ScriptWrapper::FromScript(reinterpret_cast<std::unique_ptr<Script>*>(&script))};
    auto aggr = ScriptAggregator(vec);
    while (aggr.HasNext()) {
        auto next = aggr.GetNext();
        REQUIRE(next != nullptr);
        next.As<TestScript>()->TestMethod(ran);
    }
    CHECK(ran == 1);
}

TEST_CASE("Script Aggregator properly iterates multiple scripts.") {
    auto script = std::make_unique<TestScript>("test");
    auto script2 = std::make_unique<TestScript>("test2");
    auto script3 = std::make_unique<TestScript>("test3");
    auto ran = 0;
    auto vec =
        ArbUt::List<ScriptWrapper>{ScriptWrapper::FromScript(reinterpret_cast<std::unique_ptr<Script>*>(&script)),
                                   ScriptWrapper::FromScript(reinterpret_cast<std::unique_ptr<Script>*>(&script2)),
                                   ScriptWrapper::FromScript(reinterpret_cast<std::unique_ptr<Script>*>(&script3))};
    auto aggr = ScriptAggregator(vec);
    while (aggr.HasNext()) {
        auto next = aggr.GetNext();
        REQUIRE(next != nullptr);
        next.As<TestScript>()->TestMethod(ran);
    }
    CHECK(ran == 3);
}

TEST_CASE("Script Aggregator properly iterates Script Set.") {
    Script* script = new TestScript("test");
    Script* script2 = new TestScript("test2");
    Script* script3 = new TestScript("test3");
    auto ran = 0;
    auto set = ScriptSet();
    set.Add(script);
    set.Add(script2);
    set.Add(script3);
    auto vec = ArbUt::List<ScriptWrapper>{ScriptWrapper::FromSet(&set)};
    auto aggr = ScriptAggregator(vec);
    while (aggr.HasNext()) {
        auto next = aggr.GetNextNotNull();
        REQUIRE(next != nullptr);
        next.As<TestScript>()->TestMethod(ran);
    }
    CHECK(ran == 3);
}

TEST_CASE("Script Aggregator properly iterates data of Script Set and Script.") {
    auto script = std::make_unique<TestScript>("test");
    Script* script2 = new TestScript("test2");
    Script* script3 = new TestScript("test3");
    auto ran = 0;
    auto set = ScriptSet();
    set.Add(script2);
    set.Add(script3);
    auto vec = ArbUt::List<ScriptWrapper>{
        ScriptWrapper::FromSet(&set), ScriptWrapper::FromScript(reinterpret_cast<std::unique_ptr<Script>*>(&script))};
    auto aggr = ScriptAggregator(vec);
    while (aggr.HasNext()) {
        auto next = aggr.GetNextNotNull();
        REQUIRE(next != nullptr);
        next.As<TestScript>()->TestMethod(ran);
    }
    CHECK(ran == 3);
}

TEST_CASE("Script Aggregator properly iterates data of Script and Script Set.") {
    auto script = std::make_unique<TestScript>("test");
    Script* script2 = new TestScript("test2");
    Script* script3 = new TestScript("test3");
    auto ran = 0;
    auto set = ScriptSet();
    set.Add(script2);
    set.Add(script3);
    auto vec = ArbUt::List<ScriptWrapper>{
        ScriptWrapper::FromScript(reinterpret_cast<std::unique_ptr<Script>*>(&script)), ScriptWrapper::FromSet(&set)};
    auto aggr = ScriptAggregator(vec);
    while (aggr.HasNext()) {
        auto next = aggr.GetNextNotNull();
        REQUIRE(next != nullptr);
        next.As<TestScript>()->TestMethod(ran);
    }
    CHECK(ran == 3);
}

TEST_CASE("Script Aggregator properly iterates data of Script, Script Set and Script.") {
    auto script = std::make_unique<TestScript>("test");
    Script* script2 = new TestScript("test2");
    Script* script3 = new TestScript("test3");
    auto script4 = std::make_unique<TestScript>("test4");
    auto ran = 0;
    auto set = ScriptSet();
    set.Add(script2);
    set.Add(script3);
    auto vec = ArbUt::List<ScriptWrapper>{
        ScriptWrapper::FromScript(reinterpret_cast<std::unique_ptr<Script>*>(&script)), ScriptWrapper::FromSet(&set),
        ScriptWrapper::FromScript(reinterpret_cast<std::unique_ptr<Script>*>(&script4))};
    auto aggr = ScriptAggregator(vec);
    while (aggr.HasNext()) {
        auto next = aggr.GetNextNotNull();
        REQUIRE(next != nullptr);
        next.As<TestScript>()->TestMethod(ran);
    }
    CHECK(ran == 4);
}

TEST_CASE("Script Aggregator properly iterates when empty.") {
    auto ran = 0;
    auto vec = ArbUt::List<ScriptWrapper>{};
    auto aggr = ScriptAggregator(vec);
    while (aggr.HasNext()) {
        THROW("Aggregator returned a script, but should have been empty.");
    }
    CHECK(ran == 0);
}

TEST_CASE("Script Aggregator properly iterates empty Script Set.") {
    auto ran = 0;
    auto set = ScriptSet();
    auto vec = ArbUt::List<ScriptWrapper>{ScriptWrapper::FromSet(&set)};
    auto aggr = ScriptAggregator(vec);
    aggr.Reset();
    while (aggr.HasNext()) {
        ran++;
    }
    CHECK(ran == 0);
}

#endif