#ifdef TESTS_BUILD

#include "../../../src/Battling/ScriptHandling/ScriptSource.hpp"
#include "../../../extern/doctest.hpp"

using namespace CreatureLib;
using namespace CreatureLib::Battling;

class TestScript : public Script {
private:
    ArbUt::StringView _name;

public:
    explicit TestScript(const std::string& name) : _name(name.c_str(), name.length()){};
    const ArbUt::StringView& GetName() const noexcept override { return _name; }

    void TestMethod(int& runCount) { runCount++; }
};

class ScriptSourceWithScriptPtr : public ScriptSource {
public:
    std::unique_ptr<Script> ScriptPtr = nullptr;

protected:
    size_t ScriptCount() const override { return 1; }
    void GetActiveScripts(ArbUt::List<ScriptWrapper>& scripts) override {
        scripts.Append(ScriptWrapper::FromScript(&ScriptPtr));
    }
};

class ScriptSourceWithScriptSet : public ScriptSource {
public:
    ScriptSet Set;

protected:
    size_t ScriptCount() const override { return 1; }
    void GetActiveScripts(ArbUt::List<ScriptWrapper>& scripts) override {
        scripts.Append(ScriptWrapper::FromSet(&Set));
    }
};

TEST_CASE("Script source with unset script ptr.") {
    auto source = ScriptSourceWithScriptPtr();
    auto scripts = source.GetScriptIterator();
    REQUIRE_FALSE(scripts.HasNext());
}

TEST_CASE("Script source with script ptr being set.") {
    auto source = ScriptSourceWithScriptPtr();
    source.ScriptPtr = std::make_unique<TestScript>("foobar");
    auto scripts = source.GetScriptIterator();
    auto first = scripts.GetNext();
    CHECK(first != nullptr);
}

TEST_CASE("Script source with script ptr being set after first iteration.") {
    auto source = ScriptSourceWithScriptPtr();
    auto scripts = source.GetScriptIterator();
    REQUIRE_FALSE(scripts.HasNext());
    source.ScriptPtr = std::make_unique<TestScript>("foobar");
    scripts = source.GetScriptIterator();
    auto first = scripts.GetNext();
    CHECK(first != nullptr);
}

TEST_CASE("Script source with empty script set.") {
    auto source = ScriptSourceWithScriptSet();
    auto scripts = source.GetScriptIterator();
    scripts.Reset();
    REQUIRE_FALSE(scripts.HasNext());
}

TEST_CASE("Script source with single item script set.") {
    auto source = ScriptSourceWithScriptSet();
    auto s = new TestScript("foobar");
    source.Set.Add(s);
    auto scripts = source.GetScriptIterator();
    auto first = scripts.GetNextNotNull();
    CHECK(first != nullptr);
    CHECK(first->GetName() == "foobar");
}

TEST_CASE("Script source with multiple item script set.") {
    auto source = ScriptSourceWithScriptSet();
    auto s = new TestScript("foobar");
    auto s2 = new TestScript("foobar2");
    source.Set.Add(s);
    source.Set.Add(s2);
    auto scripts = source.GetScriptIterator();
    auto first = scripts.GetNextNotNull();
    REQUIRE(first != nullptr);
    CHECK(first->GetName() == "foobar");
    auto second = scripts.GetNextNotNull();
    REQUIRE(second != nullptr);
    CHECK(second->GetName() == "foobar2");
}

#endif
